from __future__ import annotations
import sys
import os.path
import argparse
import logging
import ipaddress
from datetime import datetime
from typing import Optional as O, Any, Tuple, List, Dict, Set

logging.basicConfig(level=logging.DEBUG)

from .dazy import Item, export_items, import_items
from .wireguard import WireguardHostType, WireguardConfigFormat
from . import (
    REMOTE_COMMAND_SETS,
    WeegeeContext, WeegeeConfig,
    WeegeeHook, WeegeeHost, WeegeePublicInterface, WeegeeInterface, WeegeePeer, WeegeeConnection,
    WeegeeServer, WeegeeClient,
    WeegeeRemoteServer,
    find_interface_other_peers, config_interface,
    sync_all_interfaces,
    setup,
)

logger = logging.getLogger(setup.__module__ + '.cli')


def find_order(n: int, orders: List[Tuple[int, str]]) -> Tuple[float, str]:
    for threshold, o in reversed(orders):
        if n >= threshold:
            return n / threshold, o
    return float(n), ''

def siify(n: int) -> str:
    orders = ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
    nf, unit = find_order(n, [(1 << (10 * i), o) for (i, o) in enumerate(orders)])
    if unit:
        unit += 'i'
    return f'{round(nf, 2)} {unit}B'

def timestampify(dt: datetime) -> str:
    delta = datetime.now() - dt
    if delta.days != 0:
        future = delta.days < 0
        orders = [(1, 'day'), (7, 'week'), (30, 'month'), (365, 'year')]
        n, unit = find_order(delta.days * (-1 if future else 1), orders)
    else:
        future = delta.seconds < 0
        orders = [(1, 'second'), (60, 'minute'), (3600, 'hour')]
        n, unit = find_order(delta.seconds * (-1 if future else 1), orders)

    return f'{round(n)} {unit}{"s" if n >= 1.5 else ""} {"from now" if future else "ago"}'

def perform_get(item: Any, var: str) -> None:
    for piece in var.split('.'):
        base, *keys = piece.split('/')
        if base:
            item = getattr(item, base)
        for k in keys:
            item = item[int(k)]

    if item is None:
        print()
    elif isinstance(item, list):
        for x in item:
            print(x)
    elif isinstance(item, dict):
        for k, v in item.items():
            print(f'{k}: {v}')
    else:
        print(item)


def main():
    parser = argparse.ArgumentParser()
    parser.set_defaults(func=None, parser=parser)
    parser.add_argument('-c', '--config', help='path to configuration file')
    parser.add_argument('--yes-i-want-to-destroy-this', action='store_true', default=False)

    commands = parser.add_subparsers(title='command')

    def do_status(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> int:
        success = True
        for interface in WeegeeInterface.find_all(ctx):
            if interface.hosts:
                print(f'{interface.name}:')
            for host in interface.hosts:
                print(f'  @ {host.name}:')
                wg_peers = {p.name: p for p in host.get_peers(interface.interface_name) or []}
                for peer in find_interface_other_peers(interface.to_public()):
                    if peer.interface.public_key not in wg_peers:
                        print(f'    {peer.name}: expected but not found')
                        success = False
                    else:
                        wg_peer = wg_peers.pop(peer.interface.public_key)
                        transfer = wg_peer.get_total_transfer()
                        rx = siify(transfer[0]) if transfer else 'nothing'
                        tx = siify(transfer[1]) if transfer else 'nothing'
                        handshake = wg_peer.latest_handshake()
                        latest_handshake = timestampify(handshake) if handshake else 'never'
                        endpoint = wg_peer.latest_endpoint()
                        latest_endpoint = f'{endpoint[0]}:{endpoint[1]}' if endpoint else ''
                        print(f'    {peer.name} ({", ".join(str(x) for x in peer.interface.addresses)}, {wg_peer.name})')
                        print(f'      latest handshake {latest_handshake}{f" from {latest_endpoint}" if latest_endpoint else ""}, {tx} sent, {rx} received')
                for wg_peer in wg_peers.values():
                    print(f'    ({wg_peer.name}): found but not expected')
                    success = False
        return 0 if success else 1

    status = commands.add_parser('status', help='show interface status')
    status.set_defaults(func=do_status, parser=status)

    def do_sync(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        sync_all_interfaces(ctx)

    sync = commands.add_parser('sync', help='synchronize interfaces')
    sync.set_defaults(func=do_sync, parser=sync)

    def do_import(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> int:
        res = import_items(ctx.instance, args.file.read(), duplicates=not args.force, replacements=not args.force, deletions=not args.no_delete)
        if not res:
            return 1

        added, deleted = res
        for name in deleted:
            logger.info(f'import: {name}: deleted')
        for name in added:
            logger.info(f'import: {name}: added')

        return 0

    import_cmd = commands.add_parser('import', help='import definitions')
    import_cmd.add_argument('-n', '--no-delete', action='store_true', help='do not delete items')
    import_cmd.add_argument('-f', '--force', action='store_true', help='replace existing items')
    import_cmd.add_argument('file', type=argparse.FileType('r', encoding='utf-8'), nargs='?', default=sys.stdin)
    import_cmd.set_defaults(func=do_import, parser=import_cmd)

    def do_remote(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        if not args.host:
            config = ctx.get_config()
            args.host = config.default_remote_host or WeegeeHost.LOCAL_HOST_NAME

        commands = set()
        for c in args.command:
            if c.startswith('@'):
                c = c[1:]
                if c not in REMOTE_COMMAND_SETS:
                    parser.error(f'unknown command set: @{c}')
                commands.update(REMOTE_COMMAND_SETS[c])
            else:
                commands.add(c)

        host = WeegeeHost.load(ctx, args.host)
        interfaces = set(args.interface)
        addresses = set(args.address)
        routes = set(args.route)
        server = WeegeeRemoteServer(host.conn.conn,
            allowed_commands=commands or None,
            allowed_interfaces=interfaces or None,
            allowed_addresses=addresses or None,
            allowed_routes=routes or None,
        )
        server.serve(sys.stdin, sys.stdout)

    remote = commands.add_parser('remote', help='remote server')
    remote.add_argument('-c', '--command', type=str, action='append', default=[], help='allowed command(s) or @-prefixed command set(s)')
    remote.add_argument('-i', '--interface', type=str, action='append', default=[], help='allowed interface(s)')
    remote.add_argument('-a', '--address', type=ipaddress.ip_interface, action='append', default=[], help='allowed interface address(es)')
    remote.add_argument('-r', '--route', type=ipaddress.ip_network, action='append', default=[], help='allowed interface route(s)')
    remote.add_argument('host', help='host', nargs='?', default=None)
    remote.set_defaults(func=do_remote, parser=remote)


    # System commands

    system = commands.add_parser('system', help='system commands')
    system.set_defaults(parser=system)
    system_commands = system.add_subparsers(title='subcommand')

    def do_setup(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        setup(ctx)

    sys_setup = system_commands.add_parser('setup', help='setup system')
    sys_setup.set_defaults(func=do_setup, parser=sys_setup)

    def do_configure(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        config = ctx.get_config()

        if args.config_dir:
            config.meta_path = args.config_dir
        if args.config_dir or args.data_dir:
            config.data_path = args.data_dir or args.config_dir
        if args.log_level:
            config.log_level = args.log_level

        if args.reset_default_client_hosts:
            config.default_client_hosts = []
        for host in args.del_default_client_host:
            host_item = WeegeeHost.load(context, host).name
            if host_item in config.default_client_hosts:
                config.default_client_hosts.remove(host_item)
        for host in args.add_default_client_host:
            host_item = WeegeeHost.load(context, host).name
            if host_item not in config.default_client_hosts:
                config.default_client_hosts.append(host_item)

        if args.reset_default_server_hosts:
            config.default_server_hosts = []
        for host in args.del_default_server_host:
            host_item = WeegeeHost.load(context, host).name
            if host_item in config.default_server_hosts:
                config.default_server_hosts.remove(host_item)
        for host in args.add_default_server_host:
            host_item = WeegeeHost.load(context, host).name
            if host_item not in config.default_server_hosts:
                config.default_server_hosts.append(host_item)

        config.save(args.file)
    
    sys_configure = system_commands.add_parser('configure', help='configure defaults')
    sys_configure.add_argument('-s', '--system-file', dest='file', action='store_const', const=WeegeeConfig.GLOBAL_PATH, help=f'store to system configuration file ({WeegeeConfig.GLOBAL_PATH})')
    sys_configure.add_argument('-u', '--user-file', dest='file', action='store_const', const=WeegeeConfig.USER_PATH, help=f'store to user configuration file ({WeegeeConfig.USER_PATH})')
    sys_configure.add_argument('-l', '--local-file', dest='file', action='store_const', const=WeegeeConfig.LOCAL_PATH, help=f'store to local configuration file ({WeegeeConfig.LOCAL_PATH})')
    sys_configure.add_argument('-f', '--file', help='store to given configuration file')
    sys_configure.add_argument('-d', '--config-dir', metavar='PATH', help='config base directory (optional)')
    sys_configure.add_argument('-D', '--data-dir', metavar='PATH', help='data base directory (optional, defaults to config base directory)')
    sys_configure.add_argument('-L', '--log-level', metavar='LEVEL', help='log level', choices=('debug', 'info', 'warning', 'error', 'critical'))
    for x in ('client', 'server'):
        sys_configure.add_argument(f'--reset-default-{x}-hosts', action='store_true')
        sys_configure.add_argument(f'--add-default-{x}-host', metavar='HOST', action='append', default=[], help=f'add default host for new {x}s')
        sys_configure.add_argument(f'--del-default-{x}-host', metavar='HOST', action='append', default=[], help=f'remove default host for new {x}s')
    sys_configure.set_defaults(func=do_configure, parser=sys_configure)

    def do_migrate(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        pass

    sys_migrate = system_commands.add_parser('migrate', help='migrate system data files')
    sys_migrate.set_defaults(func=do_migrate, parser=sys_migrate)

    def do_clean(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        success = True
        while success:
            success = False

            for peer in WeegeePeer.find_all(ctx):
                if not WeegeeConnection.find_for_peers(ctx, {peer}):
                    print(f'destroying peer: {peer.name}')
                    peer.delete()
                    success = True

            for connection in WeegeeConnection.find_all(ctx):
                if len(connection.peers) <= 1:
                    print(f'destroying connection: {connection.name}')
                    connection.delete()
                    success = True

            for interface in WeegeeInterface.find_all(ctx):
                if not WeegeePeer.find_for_interfaces(ctx, {interface}):
                    print(f'destroying interface: {interface.name}')
                    interface.delete()
                    success = True

    sys_clean = system_commands.add_parser('clean', help='clean up system data files')
    sys_clean.set_defaults(func=do_clean, parser=sys_clean)


    # Common subcommands

    def add_common_subcommands(sparser: argparse._SubParsersAction) -> None:
        def do_get(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
            item = args.objtype.load(ctx, args.name)
            for var in args.var:
                perform_get(item, var)

        get = sparser.add_parser('get', help='get variable')
        get.add_argument('name', help='name')
        get.add_argument('var', nargs='*', help='variable spec')
        get.set_defaults(func=do_get, parser=get)

        def do_delete(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
            if not args.yes_i_want_to_destroy_this:
                parser.error('please pass --yes-i-want-to-destroy-this if you really want to destroy')
            item = args.objtype.load(ctx, args.name)
            item.delete()

        delete = sparser.add_parser('destroy', help='delete')
        delete.add_argument('name', help='host name')
        delete.set_defaults(func=do_delete, parser=delete)

    # Host commands

    host = commands.add_parser('host', help='host commands')
    host.set_defaults(parser=host, objtype=WeegeeHost)
    host_commands = host.add_subparsers(title='subcommand')

    def do_add_host(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        if args.auto_manage and not args.type:
            parser.error('have to specify -t/--type when --auto-manage is enabled')
        host = WeegeeHost.create(ctx,
            args.name, type=args.type, host=args.host, user=args.user, elevate_user=args.elevate_user,
            autosync=args.auto_sync, automanage=args.auto_manage,
        )
        host.save()
        host.sync(set(), auto=True)
    
    add_host = host_commands.add_parser('create', help='add host')
    add_host.add_argument('-t', '--type', type=WireguardHostType, default=WireguardHostType.Unsupported, help='host type for auto-manage operations')
    add_host.add_argument('-H', '--host', help='remote host')
    add_host.add_argument('-u', '--user', help='username')
    add_host.add_argument('-U', '--elevate_user', metavar='USER', help='username to elevate privileges to')
    add_host.add_argument('-a', '--auto-sync', action='store_true', default=False, help='whether to auto-synchronize config')
    add_host.add_argument('-A', '--auto-manage', action='store_true', default=False, help='whether to auto-synchronize interfaces')
    add_host.add_argument('name', help='host name')
    add_host.set_defaults(func=do_add_host, parser=add_host)

    add_common_subcommands(host_commands)

    def do_configure_host(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        host = WeegeeHost.load(ctx, args.name)
        if args.type is not None:
            host.type = args.type
        if args.host is not None:
            host.host = args.host
        if args.user is not None:
            host.user = args.user
        if args.elevate_user is not None:
            host.elevate_user = args.elevate_user
        if args.auto_sync is not None:
            host.autosync = args.auto_sync
        if args.auto_manage is not None:
            host.automanage = args.auto_manage

        hook_names = host.list_hooks()

        hook_reset: Dict[str, Set[str]] = {}
        hook_added: Dict[str, Dict[str, List[str]]] = {}
        hook_removed: Dict[str, Dict[str, List[str]]] = {}
        for when in ('pre', 'post'):
            hook_reset[when] = set(getattr(args, f'reset_{when}_hooks'))
            unknown = hook_reset[when] - set(hook_names)
            if unknown:
                parser.error(f'unknown hooks for --reset-{when}-hooks: {", ".join(unknown)} (see `--help` for list)')
            hook_removed[when] = {}
            for (k, v) in getattr(args, f'del_{when}_hook'):
                if k not in hook_names:
                    parser.error(f'unknown hook for --del-{when}-hook: {k} (see `--help` for list)')
                hook_removed[when].setdefault(k, []).append(v)
            hook_added[when] = {}
            for (k, v) in getattr(args, f'add_{when}_hook'):
                if k not in hook_names:
                    parser.error(f'unknown hook for --add-{when}-hook: {k} (see `--help` for list)')
                hook_added[when].setdefault(k, []).append(v)

        for hook_name in hook_names:
            hook_conf_name = f'{args.name}/{hook_name.replace("_", "-")}'
            hook_attr_name = f'{hook_name}_hooks'

            exists = WeegeeHook.exists(ctx, hook_conf_name)
            if exists:
                hook = WeegeeHook.load(ctx, hook_conf_name)
            else:
                hook = WeegeeHook.create(ctx, hook_conf_name, pre=[], post=[])

            for when in ('pre', 'post'):
                if hook_name in hook_reset[when]:
                    setattr(hook, when, [])
                    hook_reset[when].remove(hook_name)
                for cmd in hook_removed[when].pop(hook_name, []):
                    if cmd in getattr(hook, when):
                        getattr(hook, when).remove(cmd)
                for cmd in hook_added[when].pop(hook_name, []):
                    getattr(hook, when).append(cmd)

            if hook.pre or hook.post:
                hook.save()
                if hook not in host.get_hooks(hook_name).hooks:
                    getattr(host, hook_attr_name).append(hook.item)
            elif exists:
                if hook in host.get_hooks(hook_name).hooks:
                    getattr(host, hook_attr_name).remove(hook.item)
                hook.delete()

        host.save()
        host.sync(set(), auto=True)

    configure_host = host_commands.add_parser('configure', help='configure host settings')
    configure_host.add_argument('-t', '--type', type=WireguardHostType, help='host type for auto-manage operations')
    configure_host.add_argument('-H', '--host', help='remote host')
    configure_host.add_argument('-u', '--user', help='username')
    configure_host.add_argument('-U', '--elevate_user', metavar='USER', help='username to elevate privileges to')
    configure_host.add_argument('-a', '--auto-sync', action='store_true', default=None, help='auto-synchronize config')
    configure_host.add_argument('--no-auto-sync', dest='auto_sync', action='store_false', help='do not auto-synchronize config')
    configure_host.add_argument('-A', '--auto-manage', action='store_true', default=None, help='auto-synchronize interfaces')
    configure_host.add_argument('--no-auto-manage', dest='auto_manage', action='store_false', help='do not auto-synchronize interface')
    configure_hooks = configure_host.add_argument_group(title='hooks', description=f'available hooks: {", ".join(WeegeeHost.list_hooks())}')
    for when in ('pre', 'post'):
        configure_hooks.add_argument(f'--reset-{when}-hooks', metavar='HOOK', action='append', default=[], help=f'reset hook {when}-commands')
        configure_hooks.add_argument(f'--add-{when}-hook', nargs=2, metavar=('HOOK', 'CMD'), action='append', default=[], help=f'add hook {when}-command')
        configure_hooks.add_argument(f'--del-{when}-hook', nargs=2, metavar=('HOOK', 'CMD'), action='append', default=[], help=f'remove hook {when}-command')
    configure_host.add_argument('name', help='host name')
    configure_host.set_defaults(func=do_configure_host, parser=configure_host)


    # Interface commands

    interface = commands.add_parser('interface', help='interface commands')
    interface.set_defaults(parser=interface, objtype=WeegeeInterface)
    interface_commands = interface.add_subparsers(title='subcommand')

    def do_add_interface(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        interface_name = args.interface or 'wg0'
        interface = WeegeeInterface.create(ctx, args.name, interface_name,
            private_key=args.private_key, public_key=args.public_key,
            addresses=args.address, port=args.port, mtu=args.mtu,
            hosts=[WeegeeHost.load(ctx, h) for h in args.host],
            extra=args.metadata,
        )
        interface.save()
        interface.sync(set(), auto=True)

    add_interface = interface_commands.add_parser('create', help='add new interface')
    add_interface.add_argument('-H', '--host', action='append', default=[], help='interface host(s)')
    add_interface.add_argument('-a', '--address', type=ipaddress.ip_interface, action='append', default=[], help='interface address(es)')
    add_interface.add_argument('-k', '--public-key', metavar='KEY',help='public key (optional)')
    add_interface.add_argument('-K', '--private-key', metavar='KEY', help='private key (optional)')
    add_interface.add_argument('-i', '--interface', metavar='NAME', help='interface name (optional)')
    add_interface.add_argument('-p', '--port', type=int, help='listen port')
    add_interface.add_argument('-m', '--mtu', type=int, help='interface MTU')
    add_interface.add_argument('-d', '--metadata', action='append', help='metadata item(s)')
    add_interface.add_argument('name', help='interface name')
    add_interface.set_defaults(func=do_add_interface, parser=add_interface)

    add_common_subcommands(interface_commands)

    def do_conf_interface(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        if args.host:
            host = WeegeeHost.load(ctx, args.host)
        else:
            host = None
        format = args.format or (host.config_format if host else WireguardConfigFormat.WG)
        interface = WeegeeInterface.load(ctx, args.name)
        print(config_interface(interface, format, host))

    conf_interface = interface_commands.add_parser('print-config', help='show WireGuard configuration')
    conf_interface.add_argument('-H', '--host', help='target host')
    conf_interface.add_argument('-f', '--format', type=WireguardConfigFormat, help='configuration format to use')
    conf_interface.add_argument('name', help='interface name')
    conf_interface.set_defaults(func=do_conf_interface, parser=conf_interface)

    def do_connect_interface(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        left_interface = WeegeeInterface.load(ctx, args.left)
        left_peer = WeegeePeer.create(ctx, f'{args.left}-{args.right}', left_interface,
            routes=args.left_route, host=args.left_endpoint, port=args.left_port,
        )
        left_peer.save()
        right_interface = WeegeeInterface.load(ctx, args.right)
        right_peer = WeegeePeer.create(ctx, f'{args.right}-{args.left}', right_interface,
            routes=args.right_route, host=args.right_endpoint, port=args.right_port,
        )
        right_peer.save()
        connection = WeegeeConnection.create(ctx, f'{args.left}-{args.right}', [left_peer, right_peer],
            preshared_key=args.preshared_key,
        )
        connection.save()
        connection.sync(set(), auto=True)

    connect_interface = interface_commands.add_parser('connect', help='connect two interfaces')
    for i, (argname, helpname) in enumerate([('left', 'first'), ('right', 'second')]):
        connect_interface.add_argument('-R' if i else '-r', f'--{argname}-route', metavar='ROUTE', type=ipaddress.ip_network, action='append', default=[], help=f'{helpname} interface route(s)')
        connect_interface.add_argument('-E' if i else '-e', f'--{argname}-endpoint', metavar='HOST', help=f'{helpname} interface endpoint (optional)')
        connect_interface.add_argument('-P' if i else '-p', f'--{argname}-port', metavar='PORT', type=int, help=f'{helpname} interface port (optional)')
    connect_interface.add_argument('-k', '--preshared-key', metavar='KEY', help='preshared key (optional)')
    connect_interface.add_argument('left', help='first interface')
    connect_interface.add_argument('right', help='second interface')
    connect_interface.set_defaults(func=do_connect_interface, parser=connect_interface)

    def do_disconnect_interface(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        connection = WeegeeConnection.load(ctx, f'{args.left}-{args.right}')
        connection.delete()

    disconnect_interface = interface_commands.add_parser('disconnect', help='disconnect two interfaces')
    disconnect_interface.add_argument('left', help='first interface')
    disconnect_interface.add_argument('right', help='second interface')
    disconnect_interface.set_defaults(func=do_disconnect_interface, parser=disconnect_interface)


    # Peer commands

    def add_common_peer_subcommands(sparser: argparse._SubParsersAction) -> None:
        add_common_subcommands(sparser)

        def do_export(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
            peer = args.objtype.load(ctx, args.name)
            interface = peer.interface.to_public()
            print(export_items([interface.item.purify(), peer.item.purify()]))

        export = sparser.add_parser('export', help='export peer')
        export.add_argument('name', help='name')
        export.set_defaults(func=do_export, parser=export)

        def do_connect(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
            if not WeegeeConnection.exists(ctx, args.name):
                connection = WeegeeConnection.create(ctx, args.name,
                    peers=[], preshared_key=args.preshared_key,
                )
            else:
                connection = WeegeeConnection.load(ctx, args.name)
            connection.peers += [WeegeePeer.load(ctx, peer) for peer in args.peer]
            if args.preshared_key:
                connection.preshared_key = args.preshared_key
            connection.save()
            connection.sync(set(), auto=True)

        connect = sparser.add_parser('connect', help='connect peers')
        connect.add_argument('-k', '--preshared-key', metavar='NAME', help='preshared key (optional)')
        connect.add_argument('name', help='connection name')
        connect.add_argument('peer', nargs='*', help='peer names')
        connect.set_defaults(func=do_connect, parser=connect)

        def do_disconnect(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
            connection = args.objtype.load(ctx, args.name)
            peers = connection.peers
            for peer in args.peer:
                peers.remove(WeegeePeer.load(ctx, peer).item)
            connection.peers = peers
            if connection.peers:
                connection.save()
                connection.sync(set(), auto=True)
            else:
                connection.delete()

        disconnect = sparser.add_parser('disconnect', help='disconnect peers')
        disconnect.add_argument('name', help='connection name')
        disconnect.add_argument('peer', nargs='+', help='peer names')
        disconnect.set_defaults(func=do_disconnect, parser=disconnect)

        def do_get_connection(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
            connection = args.objtype.load(ctx, args.name)
            for var in args.var:
                perform_get(connection, var)

        get_connection = sparser.add_parser('get-connect', help='get connection variable')
        get_connection.add_argument('name', help='connection name')
        get_connection.add_argument('var', nargs='*', help='variable spec')
        get_connection.set_defaults(func=do_get_connection, parser=get_connection)

        def do_print_config(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
            if args.host:
                host = WeegeeHost.load(ctx, args.host)
            else:
                host = None
            format = args.format or (host.config_format if host else WireguardConfigFormat.WG)
            peer = args.objtype.load(ctx, args.name)
            print(config_interface(peer.interface.to_full(), format, host))

        print_config = sparser.add_parser('print-config', help='show WireGuard configuration')
        print_config.add_argument('-H', '--host', help='target host')
        print_config.add_argument('-f', '--format', type=WireguardConfigFormat, help='configuration format to use')
        print_config.add_argument('name', help='peer name')
        print_config.set_defaults(func=do_print_config, parser=print_config)

    peer = commands.add_parser('peer', help='peer commands')
    peer.set_defaults(parser=peer, objtype=WeegeePeer)
    peer_commands = peer.add_subparsers(title='subcommand')

    def do_add_peer(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        interface = WeegeePublicInterface.load(ctx, args.interface)
        peer = WeegeePeer.create(ctx, args.name, interface,
            routes=args.route, host=args.endpoint, port=args.port, dns=args.dns,
            extra=args.metadata,
        )
        peer.save()
        peer.sync(set(), auto=True)

    add_peer = peer_commands.add_parser('create', help='add new peer')
    add_peer.add_argument('-r', '--route', type=ipaddress.ip_network, action='append', default=[], help='peer route(s)')
    add_peer.add_argument('-n', '--dns', type=ipaddress.ip_address, action='append', default=[], help='peer DNS server(s)')
    add_peer.add_argument('-e', '--endpoint', metavar='HOST', help='peer endpoint host (optional)')
    add_peer.add_argument('-p', '--port', type=int, help='peer endpoint port (optional)')
    add_peer.add_argument('-d', '--metadata', action='append', help='metadata item(s)')
    add_peer.add_argument('name', help='peer name')
    add_peer.add_argument('interface', help='existing interface name')
    add_peer.set_defaults(func=do_add_peer, parser=add_peer)

    add_common_peer_subcommands(peer_commands)


    # Server commands

    server = commands.add_parser('server', help='server commands')
    server.set_defaults(parser=server, objtype=WeegeeServer)
    server_commands = server.add_subparsers(title='subcommand')

    def do_add_server(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        interface_name = args.interface or f'wg-{args.name}'
        interface = WeegeeInterface.create(ctx, args.name, interface_name,
            private_key=args.private_key, public_key=args.public_key,
            addresses=args.address, port=args.port, hosts=[WeegeeHost.load(ctx, name) for name in args.host or ctx.get_config().default_server_hosts],
        )
        server = WeegeeServer.create(ctx, args.name, interface, routes=args.route, host=args.endpoint, dns=args.dns, extra=args.metadata)
        server.save()
        server.sync(set(), auto=True)

    add_server = server_commands.add_parser('create', help='add new server')
    add_server.add_argument('-H', '--host', action='append', default=[], help='server host(s)')
    add_server.add_argument('-a', '--address', type=ipaddress.ip_interface, action='append', help='interface address(es)')
    add_server.add_argument('-r', '--route', type=ipaddress.ip_network, default=[], action='append', help='interface route(s) for clients')
    add_server.add_argument('-n', '--dns', type=ipaddress.ip_address, action='append', default=[], help='DNS server(s) for clients')
    add_server.add_argument('-k', '--public-key', metavar='KEY', help='public key (optional)')
    add_server.add_argument('-K', '--private-key', metavar='KEY', help='private key (optional)')
    add_server.add_argument('-i', '--interface', metavar='NAME', help='interface name (optional)')
    add_server.add_argument('-d', '--metadata', action='append', help='metadata item(s)')
    add_server.add_argument('name', help='server name')
    add_server.add_argument('endpoint', help='endpoint host')
    add_server.add_argument('port', type=int, help='listen port')
    add_server.set_defaults(func=do_add_server, parser=add_server)

    add_common_peer_subcommands(server_commands)


    # Client commands

    client = commands.add_parser('client', help='client commands')
    client.set_defaults(parser=client, objtype=WeegeeClient)
    client_commands = client.add_subparsers(title='subcommand')

    def do_add_client(parser: argparse.ArgumentParser, args: argparse.Namespace, ctx: WeegeeContext) -> None:
        server = WeegeeServer.load(ctx, args.server)
        full_name = server.get_client_name(args.name)
        interface_name = args.interface or f'wg0'
        interface = WeegeeInterface.create(ctx, full_name, interface_name,
            private_key=args.private_key, public_key=args.public_key,
            addresses=args.address, hosts=[WeegeeHost.load(ctx, name) for name in args.host or ctx.get_config().default_client_hosts],
        )
        client, connection = WeegeeClient.create(ctx, args.name, server, preshared_key=args.preshared_key, interface=interface, routes=[], extra=args.metadata)
        connection.save()
        connection.sync(set(), auto=True)

    add_client = client_commands.add_parser('create', help='add new client')
    add_client.add_argument('-H', '--host', action='append', default=[], help='client host(s)')
    add_client.add_argument('-a', '--address', type=ipaddress.ip_interface, action='append', help='interface address(es)')
    add_client.add_argument('-k', '--public-key', metavar='KEY', help='public key (optional)')
    add_client.add_argument('-K', '--private-key', metavar='KEY', help='private key (optional)')
    add_client.add_argument('-p', '--preshared-key', metavar='KEY', help='preshared key (optional)')
    add_client.add_argument('-i', '--interface', metavar='NAME', help='interface name (optional)')
    add_client.add_argument('-d', '--metadata', action='append', help='metadata item(s)')
    add_client.add_argument('server', help='server name')
    add_client.add_argument('name', help='client name')
    add_client.set_defaults(func=do_add_client, parser=add_client)

    add_common_peer_subcommands(client_commands)


    # The main buffet

    args = parser.parse_args()
    if not args.func:
        args.parser.error('a subcommand must be provided')

    config = WeegeeConfig.find()
    if args.config:
        config.load(args.config)
    context = WeegeeContext(config)
    sys.exit(args.func(args.parser, args, context))


# isn't this a tautology?
if __name__ == '__main__':
    main()
