from __future__ import annotations
from logging import getLogger

from dataclasses import dataclass, field
from typing import Optional as O, Union as U, Any, TypeVar, Type as TypeOf, Tuple, List, Dict, Set

from .dazy import Type, Meta, Item, Template
from .wireguard import (
    IPAddress, IPNetwork, IPInterface,
    WireguardHostType, WireguardHost, WireguardConfigFormat,
    WireguardPeer, WireguardConnection, which_connection,
)
from .config import WeegeeContext
from .desc import (
    WeegeeMeta,
    WEEGEE_HOOK, WEEGEE_HOST, WEEGEE_PUB_INTERFACE, WEEGEE_INTERFACE, WEEGEE_PEER, WEEGEE_CONNECTION,
    WEEGEE_INTERFACE_CONF_WG, WEEGEE_INTERFACE_CONF_WGQUICK, WEEGEE_INTERFACE_CONF_OPENWRT,
    WEEGEE_PEER_CONF_WG, WEEGEE_PEER_CONF_OPENWRT, WEEGEE_CONF_WG, WEEGEE_CONF_OPENWRT,
)

logger = getLogger(__name__)


B = TypeVar('B', bound='WeegeeBase')

@dataclass(eq=False, repr=False)
class WeegeeBase:
    BASE: WeegeeMeta = field(init=False)
    context: WeegeeContext
    item: Item
    meta: Meta = None

    def __post_init__(self) -> None:
        if not self.meta:
            self.meta = self.get_meta(self.context)
        self.item = self.item.resolve(self.meta)

    @classmethod
    def get_name(cls, name: str) -> str:
        return f'{cls.BASE.item_prefix}/{name}'

    @classmethod
    def get_meta(cls, context: WeegeeContext) -> Meta:
        return Meta.load(context.instance, cls.BASE.get_name())

    @classmethod
    def exists(cls, context: WeegeeContext, name: str, trashed: bool = False) -> bool:
        if trashed:
            return Item.restorable(context.instance, cls.get_name(name))
        else:
            return Item.exists(context.instance, cls.get_name(name))

    @classmethod
    def create(cls: TypeOf[B], context: WeegeeContext, name: str, extra: O[List[str]] = None, **kwargs) -> B:
        meta = cls.get_meta(context)
        name = cls.get_name(name)

        extra_item = None
        if extra:
            extra_item = Item.parse(context.instance, name, extra)
            for n in list(extra_item.unresolved):
                extra_item.unresolved[n] = repr(extra_item.unresolved[n])
                extra_item.resolve_type(n, Type.parse('str'))
        item = Item.make(context.instance, name, **kwargs)

        if not item.is_complete(meta):
            raise TypeError('internal error')
        if extra_item:
            item = extra_item.merge(item)
        return cls(context, item, meta)

    @classmethod
    def find_all(cls: TypeOf[B], context: WeegeeContext) -> List[B]:
        return cls.find(context, '*')

    @classmethod
    def find(cls: TypeOf[B], context: WeegeeContext, pattern: str) -> List[B]:
        objects = []
        meta = cls.get_meta(context)
        for name in Item.filter(context.instance, cls.get_name(pattern)):
            item = Item.load(context.instance, name)
            if item.is_complete(meta):
                objects.append(cls(context, item, meta))
        return objects

    @classmethod
    def load(cls: TypeOf[B], context: WeegeeContext, name: str, path: O[str] = None, restore: bool = False) -> B:
        if restore:
            item = Item.restore(context.instance, cls.get_name(name))
        elif path:
            item = Item.loadfile(None, name, path)
        else:
            item = Item.load(context.instance, cls.get_name(name))
        return cls(context, item)

    def copy(self: B) -> B:
        return self.__class__(self.context, self.item, self.meta)

    def save(self, path: O[str] = None) -> None:
        unresolved = self.item.unresolve(self.meta)
        if path:
            unresolved.savefile(path)
        else:
            unresolved.save()

    def sync(self, log: Set[str], auto: bool = False) -> bool:
        if self.item_name in log:
            return False
        log.add(self.item_name)
        return True

    def delete(self, trash: bool = False) -> None:
        if trash:
            self.item.trash()
        else:
            self.item.delete()

    @property
    def name(self) -> str:
        return self.item_name[len(self.get_name('')):]

    @property
    def item_name(self) -> str:
        return self.item.name

    def __getattr__(self, name: str) -> Any:
        if name not in self.item:
            raise AttributeError(name)
        return self.item[name]

    def __setattr__(self, name: str, value: Any) -> None:
        if 'item' in self.__dict__:
            desc = getattr(self.__class__, name, None)
            if not isinstance(desc, property) and self.item.is_type_resolved(name):
                self.item[name] = value
                return
        super().__setattr__(name, value)

    def __dir__(self) -> List[str]:
        return list(super().__dir__()) + list(self.item.values)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self.item.name == other.item.name

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.item_name))

    def __str__(self) -> str:
        return str(self.item)

    def __repr__(self) -> str:
        return str(self.item)


@dataclass(eq=False, repr=False)
class WeegeeHook(WeegeeBase):
    BASE = WEEGEE_HOOK

    @classmethod
    def create(cls, ctx: WeegeeContext, name: str, pre: List[str], post: List[str]) -> 'WeegeeHook':
        return super().create(ctx, name,
            pre=pre, post=post,
        )

@dataclass
class WeegeeHookRunner:
    hooks: List[WeegeeHook]
    conn: WireguardConnection
    kwargs: Dict[str, Any]

    def _run(self, cmd: str) -> str:
        for k, v in self.kwargs.items():
            cmd = cmd.replace('%' + k, str(v))
        return self.conn._run(cmd, shell=True)

    def __enter__(self) -> 'WeegeeHookRunner':
        for h in self.hooks:
            for cmd in h.pre:
                self._run(cmd)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if not exc_type:
            for h in self.hooks:
                for cmd in h.post:
                    self._run(cmd)

@dataclass(eq=False, repr=False)
class WeegeeHost(WeegeeBase):
    BASE = WEEGEE_HOST
    LOCAL_HOST_NAME = 'local'
    conn: WireguardHost = field(init=False)

    @classmethod
    def create(cls, ctx: WeegeeContext, name: str, type: WireguardHostType = WireguardHostType.Unsupported, host: O[str] = None, user: O[str] = None, elevate_user: O[str] = None, automanage: bool = False, autosync: bool = False, extra: O[List[str]] = None) -> 'WeegeeHost':
        return super().create(ctx, name,
            automanage=automanage,
            autosync=autosync,
            type=type.value,
            host=host,
            user=user,
            elevate_user=elevate_user,
            extra=extra,
        )

    @property
    def type(self) -> WireguardHostType:
        return WireguardHostType(self.item['type'])

    @type.setter
    def type(self, value: WireguardHostType) -> None:
        self.item['type'] = value.value

    def __post_init__(self) -> None:
        super().__post_init__()
        self.conn = WireguardHost(WeegeeHookedWireguardConnection(self, which_connection(self.type)(self.host, self.user, self.elevate_user)))

    @classmethod
    def get_local_host(cls, ctx: WeegeeContext) -> 'WeegeeHost':
        return cls.load(ctx, cls.LOCAL_HOST_NAME)

    @classmethod
    def list_hooks(cls) -> List[str]:
        return [
            'interface_add', 'interface_del',
            'address_add', 'address_del', 'address_ipv4_add', 'address_ipv4_del', 'address_ipv6_add', 'address_ipv6_del',
            'route_add', 'route_del', 'route_ipv4_add', 'route_ipv4_del', 'route_ipv6_add', 'route_ipv6_del',
            'configure',
        ]

    def get_config_format(self, name: str) -> O[WireguardConfigFormat]:
        return self.conn.conn.get_config_format(name)

    def get_hooks(self, name: str, **kwargs) -> WeegeeHookRunner:
        return WeegeeHookRunner([WeegeeHook(self.context, x) for x in getattr(self, f'{name}_hooks')], self.conn.conn, kwargs)

    def get_peers(self, name: str) -> O[List[WireguardPeer]]:
        interface = self.conn.get_interface(name)
        if not interface:
            return None
        return interface.list_peers()

    def sync_interface(self, name: str, mtu: O[int], addresses: List[IPInterface], routes: List[IPNetwork], config: str) -> bool:
        interface = self.conn.get_interface(name)
        if not interface and self.automanage:
            interface = self.conn.create_interface(name, mtu)
        if not interface:
            return False
        if self.automanage:
            interface.sync(addresses, routes, mtu)
        interface.sync_config(config)
        return True

    def delete_interface(self, name: str) -> None:
        interface = self.conn.get_interface(name)
        if interface:
            interface.clear_config()
            if self.automanage:
                interface.delete()

@dataclass
class WeegeeHookedWireguardConnection(WireguardConnection):
    host: WeegeeHost
    inner: WireguardConnection

    @staticmethod
    def _get_family(obj: U[IPAddress, IPInterface, IPNetwork]) -> str:
        return f'ipv{obj.version}'

    def __getattr__(self, name: str) -> Any:
        return getattr(self.inner, name)

    def create_interface(self, name: str) -> None:
        with self.host.get_hooks('interface_add', i=name):
            return self.inner.create_interface(name)

    def destroy_interface(self, name: str) -> None:
        with self.host.get_hooks('interface_del', i=name):
            return self.inner.destroy_interface(name)

    def add_address(self, name: str, address: IPInterface) -> None:
        family_hook = f'address_{self._get_family(address)}_add'
        with self.host.get_hooks('address_add', i=name, a=str(address)), self.host.get_hooks(family_hook, i=name, a=str(address)):
            return self.inner.add_address(name, address)

    def delete_address(self, name: str, address: IPInterface) -> None:
        family_hook = f'address_{self._get_family(address)}_del'
        with self.host.get_hooks('address_del', i=name, a=str(address)), self.host.get_hooks(family_hook, i=name, a=str(address)):
            return self.inner.delete_address(name, address)

    def delete_all_addresses(self, name: str) -> None:
        for address in self.get_addresses(name):
            self.delete_address(name, address)

    def add_route(self, name: str, route: IPNetwork) -> None:
        family_hook = f'route_{self._get_family(route)}_add'
        with self.host.get_hooks('route_add', i=name, r=str(route)), self.host.get_hooks(family_hook, i=name, r=str(route)):
            return self.inner.add_route(name, route)

    def delete_route(self, name: str, route: IPNetwork) -> None:
        family_hook = f'route_{self._get_family(route)}_del'
        with self.host.get_hooks('route_del', i=name, r=str(route)), self.host.get_hooks(family_hook, i=name, r=str(route)):
            return self.inner.delete_route(name, route)

    def delete_all_routes(self, name: str) -> None:
        for route in self.get_routes(name):
            self.delete_route(name, route)


@dataclass(eq=False, repr=False)
class WeegeePublicInterface(WeegeeBase):
    BASE = WEEGEE_PUB_INTERFACE

    def is_full(self) -> bool:
        return self.item.is_complete(WeegeeInterface.get_meta(self.context))

    def to_full(self) -> 'WeegeeInterface':
        return WeegeeInterface(self.context, self.item)

    def to_public(self) -> 'WeegeePublicInterface':
        return WeegeePublicInterface.load(self.context, self.name)

    def save(self, path: O[str] = None) -> None:
        if self.is_full():
            if type(self) != WeegeeInterface:
                self.to_full().save(path=path)
                return
        else:
            if type(self) != WeegeePublicInterface:
                self.to_public().save(path=path)
                return
        super().save(path=path)

@dataclass(eq=False, repr=False)
class WeegeeInterface(WeegeePublicInterface):
    BASE = WEEGEE_INTERFACE

    @classmethod
    def find_for_hosts(cls, ctx: WeegeeContext, hosts: Set[WeegeeHost]) -> Set['WeegeeInterface']:
        return set(i for i in cls.find_all(ctx) if set(i.hosts) & hosts)

    @classmethod
    def create(cls, ctx: WeegeeContext, name: str, interface_name: str, private_key: O[str] = None, public_key: O[str] = None, port: O[int] = None, mtu: O[int] = None, hosts: List[WeegeeHost] = [], addresses: List[IPInterface] = [], extra: O[List[str]] = None) -> 'WeegeeInterface':
        for host in [WeegeeHost.get_local_host(ctx)] + hosts:
            try:
                private_key = private_key or host.conn.gen_private_key()
                public_key = public_key or host.conn.get_public_key(private_key)
                break
            except Exception as e:
                logger.warn(f'could not generate keypair on host {host.name!r}: {e!r}')
        else:
            logger.critical('could not generate public/private keypair automatically: please pass explicitly!')
            raise ValueError()

        return super().create(ctx, name,
            hosts=[h.item for h in hosts],
            interface_name=interface_name,
            public_key=public_key, private_key=private_key,
            addresses=addresses, port=port, mtu=mtu,
            extra=extra,
        )

    def save(self, path: O[str] = None) -> None:
        for host in self.hosts:
            host.save()
        super().save(path)

    def sync(self, log: Set[str], auto: bool = False) -> bool:
        if super().sync(log, auto):
            for host in self.hosts:
                host.sync(log, auto)
            return True
        return False

    def delete(self, trash: bool = True) -> None:
        peers = WeegeePeer.find_for_interfaces(self.context, {self})
        for p in peers:
            p.delete()
        for h in self.hosts:
            if h.autosync:
                h.delete_interface(self.interface_name)
        super().delete(trash=trash)

    @property
    def hosts(self) -> List[WeegeeHost]:
        return [WeegeeHost(self.context, x) for x in self.item['hosts']]

    def gen_config(self, peers: Set['WeegeePeer'], format: WireguardConfigFormat, host: O[WeegeeHost] = None) -> str:
        if format == WireguardConfigFormat.WG:
            template_meta = WEEGEE_INTERFACE_CONF_WG
        elif format == WireguardConfigFormat.WGQuick:
            template_meta = WEEGEE_INTERFACE_CONF_WGQUICK
        elif format == WireguardConfigFormat.OpenWrt:
            template_meta = WEEGEE_INTERFACE_CONF_OPENWRT
        else:
            raise ValueError('unimplemented config format: {}'.format(format))

        template = Template.load(self.context.instance, template_meta.get_name())
        config = template_meta.make_config(self.context.instance,
            host=host.item,
            interface=self.item,
            peers={p.name: p.item for p in peers},
        )
        return template.render(config)

@dataclass(eq=False, repr=False)
class WeegeePeer(WeegeeBase):
    BASE = WEEGEE_PEER

    @classmethod
    def find_for_interfaces(cls, ctx: WeegeeContext, interfaces: Set[WeegeePublicInterface]) -> Set['WeegeePeer']:
        return set(p for p in cls.find_all(ctx) if p.interface.to_public() in interfaces)

    @classmethod
    def create(cls, ctx: WeegeeContext, name: str, interface: WeegeePublicInterface, routes: List[IPNetwork], host: O[str] = None, port: O[int] = None, dns: List[IPAddress] = [], extra: O[List[str]] = None) -> 'WeegeePeer':
        return super().create(ctx, name,
            interface=interface.item,
            routes=routes,
            host=host, port=port,
            dns=dns,
            extra=extra,
        )

    def save(self, path: O[str] = None) -> None:
        self.interface.save()
        super().save(path)

    def sync(self, log: Set[str], auto: bool = False) -> bool:
        if super().sync(log, auto):
            logger.info(f'syncing peer: {self.name}')
            if self.interface.is_full():
                sync_interface(self.interface.to_full(), auto=auto, log=log)
            return True
        return False

    def delete(self, trash: bool = False) -> None:
        for c in WeegeeConnection.find_for_peers(self.context, {self}):
            peers = c.peers
            peers.remove(self)
            c.peers = peers
            if len(c.peers) <= 1:
                c.delete(trash=trash)
            else:
                c.save()
        super().delete(trash=trash)

    @property
    def interface(self) -> WeegeePublicInterface:
        interface = WeegeePublicInterface(self.context, self.item['interface'])
        if interface.is_full():
            return interface.to_full()
        else:
            return interface

    @property
    def endpoint(self) -> O[str]:
        if not self.host:
            return None
        return f'{self.host}:{self.port or self.interface.port}'

@dataclass(eq=False, repr=False)
class WeegeeConnection(WeegeeBase):
    BASE = WEEGEE_CONNECTION

    @classmethod
    def find_for_peers(cls, ctx: WeegeeContext, peers: Set[WeegeePeer]) -> Set['WeegeeConnection']:
        return set(c for c in cls.find_all(ctx) if set(c.peers) & peers)

    @classmethod
    def create(cls, ctx: WeegeeContext, name: str, peers: List[WeegeePeer], preshared_key: O[str] = None, extra: O[List[str]] = None) -> 'WeegeeConnection':
        if not preshared_key:
            hosts = {WeegeeHost.get_local_host(ctx)}
            for p in peers:
                interface = p.interface
                if interface.is_full():
                    hosts.update(interface.hosts)
            for host in hosts:
                try:
                    preshared_key = host.conn.gen_preshared_key()
                    break
                except Exception as e:
                    logger.warn(f'could not generate preshared key on host {host.name!r}: {e!r}')
            else:
                logger.critical('could not generate preshared key automatically: please pass explicitly!')
                raise ValueError()
        return super().create(ctx, name,
            peers=[p.item for p in peers],
            preshared_key=preshared_key,
            extra=extra,
        )

    def save(self, path: O[str] = None) -> None:
        for peer in self.peers:
            peer.save()
        super().save(path)

    def sync(self, log: Set[str], auto: bool = False) -> bool:
        if super().sync(log, auto):
            for peer in self.peers:
                peer.sync(log, auto)
            return True
        return False

    @property
    def peers(self) -> List[WeegeePeer]:
        return [WeegeePeer(self.context, x) for x in self.item['peers']]

    @peers.setter
    def peers(self, value: List[WeegeePeer]) -> None:
        self.items['peers'] = [x.item for x in value]

    def gen_config(self, target: WeegeePeer, format: WireguardConfigFormat) -> str:
        peers = [p for p in self.peers if p != target]

        if format in (WireguardConfigFormat.WG, WireguardConfigFormat.WGQuick):
            template_meta = WEEGEE_PEER_CONF_WG
        elif format == WireguardConfigFormat.OpenWrt:
            template_meta = WEEGEE_PEER_CONF_OPENWRT
        else:
            raise ValueError('unimplemented config format: {}'.format(format))

        template = Template.load(self.context.instance, template_meta.get_name())
        config = template_meta.make_config(self.context.instance,
            interface=target.interface.item,
            connection=self.item,
            peers={p.name: p.item for p in peers},
        )
        return template.render(config)


def do_config_interface(interface: WeegeeInterface, format: WireguardConfigFormat, peers: Set[WeegeePeer], connections: Set[WeegeeConnection], host: O[WeegeeHost] = None) -> str:
    peer_configs = []
    other_peers = set()
    for c in connections:
        for p in set(c.peers) & peers:
            peer_configs.append(c.gen_config(p, format))
        other_peers.update(set(c.peers) - peers)
    interface_config = interface.gen_config(other_peers, format, host)

    if format in (WireguardConfigFormat.WG, WireguardConfigFormat.WGQuick):
        template_meta = WEEGEE_CONF_WG
    elif format == WireguardConfigFormat.OpenWrt:
        template_meta = WEEGEE_CONF_OPENWRT
    else:
        raise ValueError('unimplemented config format: {}'.format(format))

    template = Template.load(interface.context.instance, template_meta.get_name())
    config = template_meta.make_config(interface.context.instance,
        interface_config=interface_config,
        peer_configs=peer_configs,
    )
    return template.render(config)

def do_discover_interface(interface: WeegeePublicInterface, peers: Set[WeegeePeer], connections: Set[WeegeeConnection]) -> Tuple[List[IPNetwork], Set[WeegeePeer]]:
    routes = []
    other_peers = set()
    for c in connections:
        for p in set(c.peers) - peers:
            routes.extend(p.routes)
            routes.extend(a.network for a in p.interface.addresses)
            other_peers.add(p)
    return routes, other_peers

def find_host_interfaces(host: WeegeeHost) -> Set[WeegeeInterface]:
    return WeegeeInterface.find_for_hosts(host.context, {host})

def find_interface_connections(interface: WeegeePublicInterface) -> Tuple[Set[WeegeePeer], Set[WeegeeConnection]]:
    peers = WeegeePeer.find_for_interfaces(interface.context, {interface})
    connections = WeegeeConnection.find_for_peers(interface.context, peers)
    return peers, connections

def find_interface_other_peers(interface: WeegeePublicInterface) -> Set[WeegeePeer]:
    own_peers, connections = find_interface_connections(interface)
    _, other_peers = do_discover_interface(interface, own_peers, connections)
    return other_peers

def do_sync_interface(interface: WeegeeInterface, peers: Set[WeegeePeer], connections: Set[WeegeeConnection], auto: bool = False) -> Set[WeegeePeer]:
    routes, other_peers = do_discover_interface(interface, peers, connections)
    for host in interface.hosts:
        if auto and not host.autosync:
            continue
        config = do_config_interface(interface, host.get_config_format(interface.interface_name), peers, connections, host)
        if not host.sync_interface(interface.interface_name, interface.mtu, interface.addresses, routes, config):
            raise ValueError(f'host {host.name} failed to sync interface!')
    return other_peers

def sync_interface(interface: WeegeeInterface, auto=False, log=None) -> None:
    if log is None:
        log = set()
    if interface.item_name in log:
        return
    log.add(interface.item_name)
    logger.info(f'syncing interface: {interface.name}')

    all_peers, all_connections = find_interface_connections(interface.to_public())
    other_peers = do_sync_interface(interface, all_peers, all_connections, auto=auto)

    for p in other_peers:
        peer_interface = p.interface
        if peer_interface.is_full():
            sync_interface(peer_interface.to_full(), auto=auto, log=log)

def sync_all_interfaces(context: WeegeeContext, auto=False, log=None) -> None:
    if log is None:
        log = set()
    for interface in WeegeeInterface.find_all(context):
        sync_interface(interface, auto=auto, log=log)

def config_interface(interface: WeegeeInterface, format: WireguardConfigFormat, host: O[WeegeeHost] = None) -> str:
    peers, connections = find_interface_connections(interface.to_public())
    return do_config_interface(interface, format, peers, connections, host)
