from __future__ import annotations
from dataclasses import dataclass
from typing import Union as U, Any, List, Dict

from .dazy import Instance, Config, Type, LitType, OptionalType, RefType, ArrType, MapType, Meta


@dataclass
class WeegeeDesc:
    name: str
    version: int

    def get_name(self) -> str:
        return f'{self.name}@{self.version}'

@dataclass
class WeegeeMeta(WeegeeDesc):
    spec: List[str]
    item_prefix: str = ''


WEEGEE_HOOK = WeegeeMeta(
    name='wg/hook',
    version=1,
    spec=[
        'pre: [str] = []',
        'post: [str] = []',
    ],
    item_prefix='wg/hook',
)
WEEGEE_HOST = WeegeeMeta(
    name='wg/host',
    version=1,
    spec=[
        'autosync: bool = 0',
        'automanage: bool = 0',
        'type: ?str = ',
        'host: ?str = ',
        'user: ?str = ',
        'elevate_user: ?str = ',
        f'interface_add_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'interface_del_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'address_add_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'address_del_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'address_ipv4_add_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'address_ipv4_del_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'address_ipv6_add_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'address_ipv6_del_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'route_add_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'route_del_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'route_ipv4_add_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'route_ipv4_del_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'route_ipv6_add_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'route_ipv6_del_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
        f'configure_hooks: [@{WEEGEE_HOOK.get_name()}] = []',
    ],
    item_prefix='wg/host',
)
WEEGEE_PUB_INTERFACE = WeegeeMeta(
    name='wg/pub-intf',
    version=1,
    spec=[
        'addresses: [ipintf]',
        'public_key: str',
    ],
    item_prefix='wg/intf',
)
WEEGEE_INTERFACE = WeegeeMeta(
    name='wg/intf',
    version=1,
    spec=WEEGEE_PUB_INTERFACE.spec + [
        'private_key: str',
        f'hosts: [@{WEEGEE_HOST.get_name()}]',
        'interface_name: str',
        'port: ?int = ',
        'mtu: ?int = ',
    ],
    item_prefix=WEEGEE_PUB_INTERFACE.item_prefix,
)
WEEGEE_PEER = WeegeeMeta(
    name='wg/peer',
    version=1,
    spec=[
        f'interface: @{WEEGEE_INTERFACE.get_name()}',
        'routes: [ipnet]',
        'host: ?str = ',
        'port: ?int = ',
        'dns: [ipaddr] = []',
    ],
    item_prefix='wg/peer',
)
WEEGEE_CONNECTION = WeegeeMeta(
    name='wg/conn',
    version=1,
    spec=[
        f'peers: [@{WEEGEE_PEER.get_name()}]',
        'preshared_key: ?str = ',
    ],
    item_prefix='wg/conn',
)
WEEGEE_CONFIG = WeegeeMeta(
    name='wg/config',
    version=1,
    spec=[
        'default_server_hosts: [str] = []',
        'default_client_hosts: [str] = []',
        'default_remote_host: ?str = ',
        'log_level: str = "info"',
        'meta_path: str = "."',
        'data_path: ?str = ',
    ],
)


@dataclass
class WeegeeTemplate(WeegeeDesc):
    template: str
    variables: Dict[str, Any]

    def resolve_type(self, instance: Instance, type: Any) -> Type:
        ttype: Type
        if isinstance(type, dict):
            ktype = next(iter(type))
            vtype = type[ktype]
            ttype = MapType(self.resolve_type(instance, ktype), self.resolve_type(instance, vtype))
        elif isinstance(type, list):
            etype = type[0]
            ttype = ArrType(self.resolve_type(instance, etype))
        elif isinstance(type, WeegeeMeta):
            ttype = OptionalType(RefType(type.get_name()))
        else:
            ttype = LitType(type.__name__, type)
        return ttype

    def make_config(self, instance: Instance, **kwargs) -> 'Config':
        types = {}
        for k, v in kwargs.items():
            vtype = self.variables[k]
            ttype = self.resolve_type(instance, vtype)
            types[k] = ttype
        c = Config.make(instance, '<vars>', **kwargs)
        for k, type in types.items():
            c.resolve_type(k, type)
        return c


WEEGEE_INTERFACE_CONF_WG = WeegeeTemplate(
    name='wg/interface-conf/wg',
    version=1,
    template="""
[Interface]
{% if interface.port -%}
ListenPort = {{interface.port}}
{% endif -%}
PrivateKey = {{interface.private_key}}
""".strip(),
    variables={'host': WEEGEE_HOST, 'interface': WEEGEE_INTERFACE, 'peers': {str: WEEGEE_PEER}},
)

WEEGEE_INTERFACE_CONF_WGQUICK = WeegeeTemplate(
    name='wg/interface-conf/wg-quick',
    version=1,
    template="""
{% set dns = peers.values() | sum(attribute='dns', start=[]) %}
[Interface]
Address = {{ interface.addresses | join(", ") }}
{% if dns -%}
DNS = {{ dns | join(", ") }}
{% endif -%}
{% if interface.port -%}
ListenPort = {{interface.port}}
{% endif -%}
{% if interface.mtu -%}
MTU = {{interface.mtu}}
{% endif -%}
PrivateKey = {{interface.private_key}}

{%- if host %}
{%-   for hook in host.interface_add_hooks %}
{%-     for cmd in hook.pre %}
PreUp = {{cmd}}
{%-     endfor %}
{%-     for cmd in hook.post %}
PostUp = {{cmd}}
{%-     endfor %}
{%-   endfor %}
{%-   for hook in host.interface_del_hooks %}
{%-     for cmd in hook.pre %}
PreDown = {{cmd}}
{%-     endfor %}
{%-     for cmd in hook.post %}
PostDown = {{cmd}}
{%-     endfor %}
{%-   endfor %}
{%- endif %}
""".strip(),
    variables={'host': WEEGEE_HOST, 'interface': WEEGEE_INTERFACE, 'peers': {str: WEEGEE_PEER}},
)

WEEGEE_INTERFACE_CONF_OPENWRT = WeegeeTemplate(
    name='wg/interface-conf/openwrt',
    version=1,
    template="""
{%- set iname = interface.interface_name | to_ident -%}
delete network.{{iname}}
set network.{{iname}}="interface"
set network.{{iname}}.proto="wireguard"
set network.{{iname}}.private_key="{{interface.private_key}}"
{%- if interface.port %}
set network.{{iname}}.mtu="{{interface.mtu}}"
{%- endif %}
{%- if interface.port %}
set network.{{iname}}.listen_port={{interface.port}}
{%- endif %}
{%- for addr in interface.addresses %}
add_list network.{{iname}}.addresses="{{addr}}"
{%- endfor %}
""".strip(),
    variables={'host': WEEGEE_HOST, 'interface': WEEGEE_INTERFACE, 'peers': {str: WEEGEE_PEER}},
)

WEEGEE_PEER_CONF_WG = WeegeeTemplate(
    name='wg/peer-conf/wg',
    version=1,
    template="""
{% for name, peer in peers.items() -%}
[Peer]
# Peer: {{name}}
{%   if peer.host and (peer.interface.port or peer.port) -%}
Endpoint = {{peer.host}}:{{peer.port or peer.interface.port}}
{%   endif -%}
PublicKey = {{peer.interface.public_key}}
PresharedKey = {{connection.preshared_key}}
AllowedIPs = {{ ((peer.interface.addresses | map(attribute='network') | list) + peer.routes) | join(", ") }}
PersistentKeepalive = 30
{% endfor -%}
""".strip(),
    variables={'interface': WEEGEE_INTERFACE, 'connection': WEEGEE_CONNECTION, 'peers': {str: WEEGEE_PEER}},
)

WEEGEE_PEER_CONF_OPENWRT = WeegeeTemplate(
    name='wg/peer-conf/openwrt',
    version=1,
    template="""
{%- set iname = interface.interface_name | to_ident -%}
{%- for name, peer in peers.items() %}
{%-   set pname = interface.interface_name + "_" + name | to_ident -%}
delete network.{{pname}}
set network.{{pname}}="wireguard_{{iname}}"
set network.{{pname}}.public_key="{{peer.interface.public_key}}"
set network.{{pname}}.preshared_key="{{connection.preshared_key}}"
set network.{{pname}}.route_allowed_ips=true
{%-   if peer.host and (peer.interface.port or peer.port) %}
set network.{{pname}}.endpoint_host="{{peer.host}}"
set network.{{pname}}.endpoint_port={{peer.port or peer.interface.port}}
{%-   endif %}
{%-   for addr in peer.interface.adresses | map(attribute='network') %}
add_list network.{{pname}}.allowed_ips="{{addr}}"
{%-   endfor %}
{%-   for addr in peer.routes %}
add_list network.{{pname}}.allowed_ips="{{addr}}"
{%-   endfor %}
set network.{{pname}}.persistent_keepalive=30

{%-   for server in peer.dns %}
add_list dhcp.@dnsmasq[0].server="{{server}}"
{%-   endfor %}
{%- endfor %}
""".strip(),
    variables={'interface': WEEGEE_INTERFACE, 'connection': WEEGEE_CONNECTION, 'peers': {str: WEEGEE_PEER}},
)

WEEGEE_CONF_WG = WeegeeTemplate(
    name='wg/conf/wg',
    version=1,
    template="""
{{interface_config}}

{% for peer_config in peer_configs -%}
{{peer_config}}
{% endfor -%}
""".strip(),
    variables={'interface_config': str, 'peer_configs': [str]},
)

WEEGEE_CONF_OPENWRT = WeegeeTemplate(
    name='wg/conf/openwrt',
    version=1,
    template="""
{{interface_config}}

{% for peer_config in peer_configs -%}
{{peer_config}}

{% endfor -%}

commit network
commit dhcp
""".strip(),
    variables={'interface_config': str, 'peer_configs': [str]},
)
